/*
** Module   : Buffer.java
** Abstract : interface which must be extended by any business DMO interface
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description-----------------------
** 001 ECF 20130118 Created initial version.
*/ 

package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;

/**
 * Defines the public API for services representing Progress builtin
 * functions, methods, attributes, and language statements, which operate on
 * record buffers.
 * <p>
 * Each DMO interface should contain a public, inner interface which extends
 * both this class and the enclosing DMO interface. This enables DMO proxies
 * used by business logic to invoke DMO-specific methods, as well as methods
 * which provide standard buffer services, without requiring a cast.
 */
public interface Buffer
extends WrappedResource
{
   /**
    * Get the record ID of the current buffer's current record, if any.  P2J
    * implementation of the RECID function.
    * <p>
    * This method will return a 32-bit recid which represents the internal
    * row ID of the given record.  If the actual row ID (which is implemented
    * as a wider integer) cannot fit due to overflow, an error condition will
    * be raised.
    *
    * @return  Record ID of the underlying data record. If no record currently
    *          backs the buffer associated with <code>dmo</code>, a recid
    *          initialized as unknown value is returned.
    * 
    * @throws  ErrorConditionException
    *          if the actual record ID is too wide to fit within a 32-bit
    *          integer.
    */
   public recid recordID();
   
   /**
    * Get the row ID of the current buffer's current record, if any.  P2J
    * implementation of the ROWID function.
    *
    * @return  Row ID of the underlying data record.  If no record currently
    *          backs the buffer associated with <code>dmo</code>, a
    *          <code>rowid</code> initialized as unknown value is returned.
    */
   public rowid rowID();
   
   /**
    * Indicate whether a backing data record currently is available.
    *
    * @return  <code>true</code> if data is available, <code>false</code> if
    *          no record backs this buffer.
    */
   public logical available();
   
   /**
    * Indicate whether a backing data record currently is available.
    *
    * @return  <code>true</code> if data is available, <code>false</code> if
    *          no record backs this buffer.
    */
   public boolean _available();
   
   /**
    * Checks if the record in the buffer has changed in database since the 
    * last <code>FIND CURRENT</code> or <code>GET CURRENT</code>.
    * P2J implementation of 4GL function <code>CURRENT-CHANGED</code>.
    *
    * @return  <code>true</code> if the database record differs from the
    *          value in memory buffer that was queried with the last
    *          <code>FIND</code> and <code>false</code> otherwise
    */
   public logical currentChanged();
   
   /**
    * Checks if the record in the buffer has changed in database since the 
    * last <code>FIND CURRENT</code> or <code>GET CURRENT</code>.
    * P2J implementation of 4GL function <code>CURRENT-CHANGED</code>.
    *
    * @return  <code>true</code> if the database record differs from the
    *          value in memory buffer that was queried with the last
    *          <code>FIND</code> and <code>false</code> otherwise
    */
   public boolean _currentChanged();
   
   /**
    * Indicate whether the buffer associated currently contains a record, and
    * whether that record is transient.  A record is transient if it has been
    * newly created and has not yet been saved to the database.
    *
    * @return  <code>true</code> if a record exists and is new, else
    *          <code>false</code>.
    */
   public logical newlyCreated();
   
   /**
    * Indicate whether the buffer associated currently contains a record, and
    * whether that record is transient.  A record is transient if it has been
    * newly created and has not yet been saved to the database.
    *
    * @return  <code>true</code> if a record exists and is new, else
    *          <code>false</code>.
    */
   public boolean _newlyCreated();
   
   /**
    * Indicate whether the last record retrieval attempt was denied because
    * the target record was locked.
    *
    * @return  <code>true</code> if the record was locked;  <code>false</code>
    *          if the attempt succeeded, or if this query has not yet
    *          attempted to retrieve a record.
    */
   public logical locked();
   
   /**
    * Indicate whether the last record retrieval attempt was denied because
    * the target record was locked.
    *
    * @return  <code>true</code> if the record was locked;  <code>false</code>
    *          if the attempt succeeded, or if this query has not yet
    *          attempted to retrieve a record.
    */
   public boolean _locked();
   
   /**
    * Indicate whether the last record retrieval attempt failed because the
    * query criteria was ambiguous.
    *
    * @return  <code>true</code> if the query was ambiguous;
    *          <code>false</code> if the attempt succeeded, or if no attempt
    *          has yet been made.
    */
   public logical ambiguous();
   
   /**
    * Indicate whether the last record retrieval attempt failed because the
    * query criteria was ambiguous.
    *
    * @return  <code>true</code> if the query was ambiguous;
    *          <code>false</code> if the attempt succeeded, or if no attempt
    *          has yet been made.
    */
   public boolean _ambiguous();
   
   /**
    * Create a new record in the buffer.
    *
    * @throws  ErrorConditionException
    *          if there is an error retrieving the next available primary key
    *          from the database or locking the record.
    */
   public void create();
   
   /**
    * Delete the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if no record currently is loaded in the buffer, or if there is
    *          an error deleting the current record from the database.
    */
   public void delete();
   
   /**
    * Delete all records associated with this buffer, which match the criteria
    * specified by the given where clause.  Note that this where clause must
    * not contain DMO alias qualifiers.
    * <p>
    * <b>NOTE:</b>  the current implementation must not be invoked for a DMO
    * which is backed by a permanent table;  only temp-table bulk delete is
    * supported at this time.
    * 
    * @param   where
    *          An HQL where clause snippet which defines the restriction
    *          criteria to apply to the delete.  All references to properties
    *          in a DMO must be unqualified.
    * @param   args
    *          Query substitution parameters required by the where clause.
    * 
    * @throws  ErrorConditionException
    *          if there is an error deleting the records.
    * @throws  UnsupportedOperationException
    *          if invoked on a buffer associated with a permanent table.
    */
   public void delete(String where, Object ...args);
   
   /**
    * Bulk delete all records associated with the buffer.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the bulk delete operation.
    */
   public void deleteAll();
   
   /**
    * Validate and release the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   public void release();
   
   /**
    * Validate the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   public void validate();
   
   /**
    * Open a read-only buffer scope (for trigger support).
    */
   public void openReadOnlyScope();
   
   /**
    * P2j Implementation of BUFFER-RELEASE method.
    * 
    * @return  TBA: probably true on success
    */
   public logical bufferRelease();

   /**
    * Get a field handle using its index.
    * P2j Implementation of BUFFER-FIELD method.
    * 
    * @param   n
    *          The index of the field.
    * 
    * @return  A handle to requested field.
    */
   public handle getField(NumberType n);

   /**
    * Obtain current value of a buffer-field object.
    * P2j Implementation of BUFFER-VALUE read method.
    * 
    * @return  The actual value from the buffer object.
    */
   public BaseDataType getValue(); 

   /**
    * Obtain current value of a buffer-field object from a extend field.
    * P2j Implementation of BUFFER-VALUE read method.
    * 
    * @param   i
    *          The index on the extend field.
    * 
    * @return  The actual value from the buffer object.
    */
   public BaseDataType getValue(int i);

   /**
    * Obtain current value of a buffer-field object from a extend field.
    * P2j Implementation of BUFFER-VALUE read method.
    * 
    * @param   i
    *          The index on the extend field.
    * 
    * @return  The actual value from the buffer object.
    */
   public BaseDataType getValue(NumberType i);

   /**
    * Sets the value of a buffer-field object from a extend field.
    * P2j Implementation of BUFFER-VALUE read method.
    * 
    * @param   i
    *          The index on the extend field.
    * @param   value
    *          The new value to be assigned to buffer.
    */
   public void setValue(int i, BaseDataType value);

   /**
    * Sets the value of a buffer-field object from a extend field.
    * P2j Implementation of BUFFER-VALUE read method.
    * 
    * @param   i
    *          The index on the extend field.
    * @param   value
    *          The new value to be assigned to buffer.
    */
   public void setValue(NumberType i, BaseDataType value);

   /**
    * Sets the value of a buffer-field object.
    * P2j Implementation of BUFFER-VALUE read method.
    * 
    * @param   value
    *          The new value to be assigned to buffer.
    */
   public void setValue(BaseDataType value);

   /**
    * Obtain ADM-DATA attribute (arbitrary string value associated with 
    * a persistent procedure.)
    * 
    * @return  ADM-DATA attribute value
    */
   public character getAdmData();

   /**
    * Sets ADM-DATA attribute (arbitrary string value associated with 
    * a persistent procedure.)
    * 
    * @param   val
    *          New ADM-DATA attribute value.
    */
   public void setAdmData(character val);

   /**
    * Obtain the number of elements in an array field.
    * 
    * @return  The number of elements in an array field, 0 for a field that 
    *          is not an array.
    */
   public integer getExtent();

   /**
    * Obtain the position of this buffer-field within the database record.
    * @return  The position within the database record.
    */
   public integer getPosition();

   /**
    * Obtain the handle of a temp-table object, if any, associated with the 
    * buffer object or the Unknown value (?) if the buffer is not associated 
    * with a temp-table object.
    * 
    * @return  The handle to associated temp-table, if any.
    */
   public handle getTableHandle();

   /**
    * Creates a new record, reset sets fields to their default values, and 
    * make a copy of the record into new the buffer.
    * This is the equivalent of BUFFER-CREATE() method of 4GL.
    * 
    * @return  TBA: probably true for success
    */
   public logical createBuffer();

   
   /**
    * Duplicated fields, determined by name, data type, and extent-matching, 
    * from the hBuff source buffer. Fields that are not common are ignored.
    * This is the equivalent of BUFFER-COPY() method of 4GL. At this moment 
    * only the simplest form is declared.
    * 
    * @param   hBuff
    *          The source to copy the fields from.
    * @return  TBA: probably true for success
    */
   public logical copyBuffer(handle hBuff);

   /**
    * Obtain the number of fields defined in the buffer’s table.
    * This is the equivalent of NUM-FIELDS attribute of 4GL.
    * 
    * @return  The number of fields defined in the buffer’s table.
    */
   public integer getNumFields();






}
