/*
** Module   : BufferImpl.java
** Abstract : implementation of Buffer interface
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------Description------------------------
** 001 ECF 20130118 Created initial version. Implementation of the Buffer
**                  interface; provides services for buffer-related, Progress
**                  methods, attributes, and some language statements.
** 002 OM  20130117 Created stub functions for CREATE-BUFFER statement.
*/ 

package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;

/**
 * This class implements services representing Progress builtin functions,
 * methods, attributes, and language statements, which operate on record
 * buffers.
 * <p>
 * Converted business logic invokes these methods on DMO proxies via a
 * DMO-specific interface in the form <code>&lt;DMO interface&gt;.Buf</code>.
 * Each such interface extends both a custom interface of DMO-specific getter
 * and setter methods, as well as the {@link Buffer} interface, which defines
 * the API for these standard services.
 * <p>
 * Dynamic proxy classes for each DMO type extend this class, so that business
 * logic can invoke these standard services directly as method calls on DMO
 * proxy instances.
 * <p>
 * This class is intentionally package private, since it is not meant to be
 * extended outside of this package (and within this package only by dynamic
 * proxy class), nor instantiated directly.
 * <p>
 * This class implements {@link BufferReference} in order to grant its
 * instance methods access to the {@link RecordBuffer} object associated with
 * each instance of the DMO proxy class which extends this class. This access
 * is necessary for these methods to do their work. Even though {@link
 * #buffer()} is a public method, external packages will not be able to invoke
 * this method (except through reflection). This is by design, as business
 * logic is not meant to have access to <code>RecordBuffer</code> instances.
 * The default implementation of this method here throws an exception, as it
 * is never intended to be called, but it needs to be present in order for
 * this class to compile, since it is referenced by other methods. Each DMO
 * proxy class overrides this method to provide a working implementation.
 */
public abstract class BufferImpl
implements Buffer
{
   /**
    * Get the {@link RecordBuffer} instance associated with this object.
    * <p>
    * This implementation exists only to satisfy the Java compiler. This
    * method is only meaningful when a proxy DMO class is created by {@link
    * RecordBuffer}, which will implement this method to return an instance
    * of <code>RecordBuffer</code> associated with the proxy DMO instance.
    *
    * @return  Record buffer associated with this object.
    * 
    * @throws  UnsupportedOperationException
    *          always.
    */
   protected abstract RecordBuffer buffer();

   /**
    * Reports if this object is valid for use.  
    *
    * @return  <code>true</code> if we are valid (can be used).
    */
   public boolean isValid()
   {
      // TODO:  implement correctly
      return true;
   }
   
   /**
    * Reports if this object is unknown.  
    *
    * @return  <code>true</code> if object is unknown.
    */
   public boolean isUnknown()
   {
      // TODO:  implement correctly
      return false;
   }
   
  /**
   * Get the record ID of the current buffer's current record, if any.  P2J
   * implementation of the RECID function.
   * <p>
   * This method will return a 32-bit recid which represents the internal
   * row ID of the given record.  If the actual row ID (which is implemented
   * as a wider integer) cannot fit due to overflow, an error condition will
   * be raised.
   *
   * @return  Record ID of the underlying data record. If no record currently
   *          backs the buffer associated with <code>dmo</code>, a recid
   *          initialized as unknown value is returned.
   * 
   * @throws  ErrorConditionException
   *          if the actual record ID is too wide to fit within a 32-bit
   *          integer.
   */
  public recid recordID()
  {
     return buffer().recordID();
  }
  
  /**
   * Get the row ID of the current buffer's current record, if any.  P2J
   * implementation of the ROWID function.
   *
   * @return  Row ID of the underlying data record.  If no record currently
   *          backs the buffer associated with <code>dmo</code>, a
   *          <code>rowid</code> initialized as unknown value is returned.
   */
  public rowid rowID()
  {
     return buffer().rowID();
  }
  
   /**
    * Indicate whether a backing data record currently is available.
    *
    * @return  <code>true</code> if data is available, <code>false</code> if
    *          no record backs this buffer.
    */
   public logical available()
   {
      return (new logical(_available()));
   }
   
   /**
    * Indicate whether a backing data record currently is available.
    *
    * @return  <code>true</code> if data is available, <code>false</code> if
    *          no record backs this buffer.
    */
   public boolean _available()
   {
      return buffer().isAvailable();
   }
   
   /**
    * Checks if the record in the buffer has changed in database since the 
    * last <code>FIND CURRENT</code> or <code>GET CURRENT</code>.
    * P2J implementation of 4GL function <code>CURRENT-CHANGED</code>.
    *
    * @return  <code>true</code> if the database record differs from the
    *          value in memory buffer that was queried with the last
    *          <code>FIND</code> and <code>false</code> otherwise
    */
   public logical currentChanged()
   {
      return new logical(_currentChanged());
   }
   
   /**
    * Checks if the record in the buffer has changed in database since the 
    * last <code>FIND CURRENT</code> or <code>GET CURRENT</code>.
    * P2J implementation of 4GL function <code>CURRENT-CHANGED</code>.
    *
    * @return  <code>true</code> if the database record differs from the
    *          value in memory buffer that was queried with the last
    *          <code>FIND</code> and <code>false</code> otherwise
    */
   public boolean _currentChanged()
   {
      return buffer().isRecordChanged();
   }
   
   /**
    * Indicate whether the buffer associated currently contains a record, and
    * whether that record is transient.  A record is transient if it has been
    * newly created and has not yet been saved to the database.
    *
    * @return  <code>true</code> if a record exists and is new, else
    *          <code>false</code>.
    */
   public logical newlyCreated()
   {
      return (new logical(_newlyCreated()));
   }
   
   /**
    * Indicate whether the buffer associated currently contains a record, and
    * whether that record is transient.  A record is transient if it has been
    * newly created and has not yet been saved to the database.
    *
    * @return  <code>true</code> if a record exists and is new, else
    *          <code>false</code>.
    */
   public boolean _newlyCreated()
   {
      return buffer().isNewlyCreated();
   }
   
   /**
    * Indicate whether the last record retrieval attempt was denied because
    * the target record was locked.
    *
    * @return  <code>true</code> if the record was locked;  <code>false</code>
    *          if the attempt succeeded, or if this query has not yet
    *          attempted to retrieve a record.
    */
   public logical locked()
   {
      return (new logical(_locked()));
   }
   
   /**
    * Indicate whether the last record retrieval attempt was denied because
    * the target record was locked.
    *
    * @return  <code>true</code> if the record was locked;  <code>false</code>
    *          if the attempt succeeded, or if this query has not yet
    *          attempted to retrieve a record.
    */
   public boolean _locked()
   {
      return buffer().wasLocked();
   }
   
   /**
    * Indicate whether the last record retrieval attempt failed because the
    * query criteria was ambiguous.
    *
    * @return  <code>true</code> if the query was ambiguous;
    *          <code>false</code> if the attempt succeeded, or if no attempt
    *          has yet been made.
    */
   public logical ambiguous()
   {
      return (new logical(_ambiguous()));
   }
   
   /**
    * Indicate whether the last record retrieval attempt failed because the
    * query criteria was ambiguous.
    *
    * @return  <code>true</code> if the query was ambiguous;
    *          <code>false</code> if the attempt succeeded, or if no attempt
    *          has yet been made.
    */
   public boolean _ambiguous()
   {
      return buffer().wasAmbiguous();
   }
   
   /**
    * Create a new record in the buffer.
    *
    * @throws  ErrorConditionException
    *          if there is an error retrieving the next available primary key
    *          from the database or locking the record.
    */
   public void create()
   {
      buffer().create();
   }
   
   /**
    * Delete the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if no record currently is loaded in the buffer, or if there is
    *          an error deleting the current record from the database.
    */
   public void delete()
   {
      try
      {
         buffer().delete();
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Delete all records associated with this buffer, which match the criteria
    * specified by the given where clause.  Note that this where clause must
    * not contain DMO alias qualifiers.
    * <p>
    * <b>NOTE:</b>  the current implementation must not be invoked for a DMO
    * which is backed by a permanent table;  only temp-table bulk delete is
    * supported at this time.
    * 
    * @param   where
    *          An HQL where clause snippet which defines the restriction
    *          criteria to apply to the delete.  All references to properties
    *          in a DMO must be unqualified.
    * @param   args
    *          Query substitution parameters required by the where clause.
    * 
    * @throws  ErrorConditionException
    *          if there is an error deleting the records.
    * @throws  UnsupportedOperationException
    *          if invoked on a buffer associated with a permanent table.
    */
   public void delete(String where, Object ...args)
   {
      try
      {
         buffer().delete(where, args);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Bulk delete all records associated with the buffer.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the bulk delete operation.
    */
   public void deleteAll()
   {
      buffer().deleteAll();
   }
   
   /**
    * Validate and release the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   public void release()
   {
      buffer().release();
   }
   
   /**
    * Validate the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   public void validate()
   {
      buffer().validate();
   }
   
   /**
    * Open a read-only buffer scope (for trigger support).
    */
   public void openReadOnlyScope()
   {
      // TODO:  implement
   }

   /**
    * Implementation of first form of CREATE-BUFFER that uses the table-name.
    * The buffer is assigned to the handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h,
                             String table,
                             String bufferName,
                             String widgetPool)
   {
      // TODO implement method body in runtime

      // just a suggestion for implementation, add the widgetPool param, too:
      // Class dmo = null; // dynamically obtain class
      // String database = null;
      // if (bufferName == null || "".equals(bufferName)
      // {
      // bufferName = generateUniqueName();
      // }
      //
      // DataModelObject buf = RecordBuffer.define(dmo, database, bufferName);
      // RecordBuffer.openScope(bufferName);
      // h.assign(bufferName);
   }

   /**
    * Implementation of first form of CREATE-BUFFER that uses the table-name.
    * The buffer is assigned to the handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle    h,
                             String    table,
                             String    bufferName,
                             character widgetPool)
   {
      BufferImpl.create(h, table, bufferName, widgetPool.getValue());
   }

   /**
    * Implementation of first form of CREATE-BUFFER that uses the table-name.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle    h,
                             character table,
                             character bufferName,
                             character widgetPool)
   {
      // TODO: check for unknown values
      BufferImpl.create(h,
                        table.getValue(),
                        bufferName.getValue(),
                        widgetPool.getValue());
   }

   /**
    * Implementation of first form of CREATE-BUFFER that uses the table-name.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    */
   public static void create(handle h, String table, String bufferName)
   {
      BufferImpl.create(h, table, bufferName, "");
   }

   /**
    * Implementation of first form of CREATE-BUFFER that uses the table-name.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    */
   public static void create(handle h, character table, character bufferName)
   {
      // TODO: check for unknown values
      BufferImpl.create(h, table.getValue(), bufferName.getValue(), "");
   }

   /**
    * Implementation of first form of CREATE-BUFFER that uses the table-name.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    */
   public static void create(handle    h, 
                             String    table, 
                             character bufferName)
   {
      String buffStr = bufferName.isUnknown() ? "" : bufferName.getValue();
      BufferImpl.create(h, table, buffStr, "");
   }

   /**
    * Implementation of first form of CREATE-BUFFER that uses the table-name.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    */
   public static void create(handle h, String table)
   {
      BufferImpl.create(h, table, "", "");
   }

   /**
    * Implementation of first form of CREATE-BUFFER that uses the table-name.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    */
   public static void create(handle h, character table)
   {
      if (table.isUnknown())
      {
         BufferImpl.create(h, table.getValue(), "", "");
      }
      // TODO: else throw some exception ?
   }

   /**
    * Implementation of 2nd form of CREATE-BUFFER that uses a temporary table
    * reference.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    */
   public static void create(handle h, Temporary tt)
   {
      // TODO: check for unknown values
      BufferImpl.create(h, tt, "", "");
   }

   /**
    * Implementation of 2nd form of CREATE-BUFFER that uses a temporary table
    * reference.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    */
   public static void create(handle h, Temporary tt, String bufferName)
   {
      // TODO: check for unknown values
      BufferImpl.create(h, tt, bufferName, "");
   }

   /**
    * Implementation of 2nd form of CREATE-BUFFER that uses a temporary table
    * reference.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    */
   public static void create(handle h, Temporary tt, character bufferName)
   {
      // TODO: check for unknown values
      BufferImpl.create(h, tt, bufferName.getValue(), "");
   }

   /**
    * Implementation of 2nd form of CREATE-BUFFER that uses a temporary table
    * reference.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle    h,
                             Temporary tt,
                             String    bufferName,
                             String    widgetPool)
   {
      // TODO write here the implementation for all overloads of
      // BufferImpl.create(handle, handle, ...)
   }

   /**
    * Implementation of 2nd form of CREATE-BUFFER that uses a temporary table
    * reference.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle    h,
                             Temporary tt,
                             character bufferName,
                             String    widgetPool)
   {
      // TODO: check for unknown values
      BufferImpl.create(h, tt, bufferName.getValue(), widgetPool);
   }

   /**
    * Implementation of 2nd form of CREATE-BUFFER that uses a temporary table
    * reference.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle    h,
                             Temporary tt,
                             String    bufferName,
                             character widgetPool)
   {
      // TODO: check for unknown values
      BufferImpl.create(h, tt, bufferName, widgetPool.getValue());
   }

   /**
    * Implementation of 2nd form of CREATE-BUFFER that uses a temporary table
    * reference. The buffer is assigned to the handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle    h,
                             Temporary tt,
                             character bufferName,
                             character widgetPool)
   {
      // TODO: check for unknown values
      BufferImpl.create(h, tt, bufferName.getValue(), widgetPool.getValue());
   }

   /**
    * Implementation of 3rd form of CREATE-BUFFER that uses a another buffer
    * already assigned to a handle.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    */
   public static void create(handle h, handle hbuf)
   {
      BufferImpl.create(h, hbuf, "", "");
   }

   /**
    * Implementation of 3rd form of CREATE-BUFFER that uses a another buffer
    * already assigned to a handle. The new buffer is assigned to the handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    */
   public static void create(handle h, handle hbuf, String bufferName)
   {
      BufferImpl.create(h, hbuf, bufferName, "");
   }

   /**
    * Implementation of 3rd form of CREATE-BUFFER that uses a another buffer
    * already assigned to a handle.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    */
   public static void create(handle    h,
                             handle    hbuf, 
                             character bufferName)
   {
      // TODO: check for unknown values
      BufferImpl.create(h, hbuf, bufferName.getValue(), "");
   }

   /**
    * Implementation of 3rd form of CREATE-BUFFER that uses a another buffer
    * already assigned to a handle. The new buffer is assigned to the handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h,
                             handle hbuf,
                             String bufferName,
                             String widgetPool)
   {
      // TODO write here the implementation for all overloads of
      // BufferImpl.create(handle, handle, ...)
   }

   /**
    * Implementation of 3rd form of CREATE-BUFFER that uses a another buffer
    * already assigned to a handle. 
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle    h,
                             handle    hbuf,
                             character bufferName,
                             String    widgetPool)
   {
      // TODO: check for unknown values
      BufferImpl.create(h, hbuf, bufferName.getValue(), widgetPool);
   }

   /**
    * Implementation of 3rd form of CREATE-BUFFER that uses a another buffer
    * already assigned to a handle.
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle    h,
                             handle    hbuf,
                             String    bufferName,
                             character widgetPool)
   {
      // TODO: check for unknown values
      BufferImpl.create(h, hbuf, bufferName, widgetPool.getValue());
   }

   /**
    * Implementation of 3rd form of CREATE-BUFFER that uses a another buffer
    * already assigned to a handle. 
    * The newly dynamically created buffer is assigned to handle h.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    * @param   bufferName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle    h,
                             handle    hbuf,
                             character bufferName,
                             character widgetPool)
   {
      // TODO: check for unknown values
      BufferImpl.create(h, hbuf, bufferName.getValue(), widgetPool.getValue());
   }

   /**
    * P2j Implementation of BUFFER-RELEASE method.
    * 
    * @return  unknown probably true for success 
    */
   public logical bufferRelease()
   {
      //TODO: implement this
      release();
      return new logical(); 
   }

   /**
    * Checks if an error occurred during a  operation on the specified 
    * temp-table buffer.
    * 
    * @param   buffName
    *          The name of a temp-table buffer.
    *
    * @return  true on buffer error
    */
   public static logical getBufferError(character buffName)
   {
      // TODO: look-up named buffer, return its ERROR attribute
      return new logical();
   }
}
